# --------------------------------------------------------------------------------------
# File: "topic_model_003_output.py"
# Dir: "topic_model_scripts\topic_model_003_output.py"
# Created: 2021-08-17
# --------------------------------------------------------------------------------------

"""
this file is meant to create output files for the topic model created from the data provided

this page was very useful here:
https://stackoverflow.com/questions/43357247/get-document-topics-and-get-term-topics-in-gensim
"""

from SandsPythonFunctions.TimerFunctions import function_timer as timer
from pathlib import Path
from tqdm import tqdm
import SandsPythonFunctions.ParquetFunctions as pf
import msgpack
import numpy as np
import pandas as pd
import pretty_errors
import tomotopy as tp


def create_pyLDAVis(mdl, pyvs_path):
    import pyLDAvis
    import numpy as np

    if not pyvs_path.exists():
        topic_term_dists = np.stack([mdl.get_topic_word_dist(k) for k in range(mdl.k)])
        doc_topic_dists = np.stack([doc.get_topic_dist() for doc in mdl.docs])
        doc_topic_dists /= doc_topic_dists.sum(axis=1, keepdims=True)
        doc_lengths = np.array([len(doc.words) for doc in mdl.docs])
        vocab = list(mdl.used_vocabs)
        term_frequency = mdl.used_vocab_freq
        vis = pyLDAvis.prepare(
            topic_term_dists,
            doc_topic_dists,
            doc_lengths,
            vocab,
            term_frequency,
            start_index=0,  # tomotopy starts topic ids with 0, pyLDAvis with 1
            sort_topics=False,  # IMPORTANT: otherwise topic_ids between pyLDAvis and tomotopy will not match
        )
        pyLDAvis.save_html(vis, open(str(pyvs_path.resolve()), "w"))


def create_topic_summery_text_page(mdl, topt_path):
    if not topt_path.exists():
        extractor = tp.label.PMIExtractor(
            min_cf=10, min_df=5, max_len=5, max_cand=10000, normalized=True
        )
        cands = extractor.extract(mdl)
        labeler = tp.label.FoRelevance(mdl, cands, min_df=5, smoothing=1e-2, mu=0.25)
        output_string = ""
        for k in range(mdl.k):
            output_string = f"{output_string}== Topic #{k} ==\n"
            output_string = f"{output_string}\nLabels: {', '.join(label for label, score in labeler.get_topic_labels(k, top_n=5))}\n"
            for word, prob in mdl.get_topic_words(k, top_n=30):
                if len(word) < 4:
                    output_string = f"{output_string}\n{word}\t\t\t{prob}"
                elif len(word) > 7:
                    output_string = f"{output_string}\n{word}\t{prob}"
                else:
                    output_string = f"{output_string}\n{word}\t\t{prob}"
            output_string = f"{output_string}\n\n"
        topt_path.write_text(output_string)


@timer
def extract_dominant_topic(mdl, ltxp_path, dtdx_path, corp_path, dtdp_path, dtcx_path):
    if not (dtdp_path.exists() and dtdx_path.exists() and dtcx_path.exists()):
        dta = pf.read_parquet_by_path(ltxp_path)
        # print(f"len dta: {len(dta)}; len mdl.docs: {len(mdl.docs)}")  # TESTCODE:
        if len(mdl.docs) == len(dta):
            corpus = tp.utils.Corpus.load(str(corp_path))
            inferred_corpus, ll = mdl.infer(corpus)
            dom_dist = []
            for doc in tqdm(inferred_corpus, desc="creating individual post dominant topic output"):
                dom_dist.append(doc.get_topic_dist().tolist())
            # print(f"type(dom_dist[0]: {type(dom_dist[0])}")  # TESTCODE:
            dta["dom_dist"] = dom_dist
        else:
            lemmatized_texts = dta["lemmatized_texts"].tolist()
            dom_dist = []
            for lemmatized_text in tqdm(lemmatized_texts, desc="Processing dominant topic by post"):
                doc = lemmatized_text.split()
                try:
                    doc = mdl.make_doc(doc)
                    topic_dist, ll = mdl.infer(doc)
                    dom_dist.append(topic_dist.tolist())
                except:
                    dom_dist.append(np.nan)
            dta["dom_dist"] = dom_dist
            dta = dta.dropna()
            dom_dist = dta["dom_dist"].tolist()
        dta["dom_topic"] = [(dist.index(max(dist)) + 1) for dist in dom_dist]
        dta["dom_topic_score"] = [max(dist) for dist in dom_dist]
        dta = dta[["id", "text", "dom_dist", "dom_topic", "dom_topic_score"]]
        pf.save_dataframe_as_parquet(dta, dtdp_path)
        print(dta.columns)
        dta.to_excel(dtdx_path)
        dta = dta.groupby("dom_topic")["id"].count()
        dta.to_excel(dtcx_path)


def create_dominant_topic_dta(domt_path, dta, text_dom_topics, percents_dom_topic):
    dta["dominant_topic"] = text_dom_topics
    dta["percent_dominant_topic"] = percents_dom_topic
    pf.save_dataframe_as_parquet(dta[["id", "dominant_topic", "percent_dominant_topic"]], domt_path)
    dta["percent_dominant_topic_string"] = [
        str(round(per, 3)).ljust(5, "0") for per in percents_dom_topic
    ]
    dta["dominant_topic_string"] = [str(dom).zfill(3) for dom in text_dom_topics]
    return dta


def create_dominant_topic_text_file(dta, dtxt_path, tpex_path, tgex_path):
    dta["text_no_newline"] = dta["text"].str.replace("\n", " |NEWLINE_CHAR| ")
    dta.to_excel(tpex_path, sheet_name="dominant_topics")
    dta = dta.sort_values(
        ["dominant_topic_string", "percent_dominant_topic_string"], ascending=[True, False]
    )
    dta["output_string"] = (
        dta["dominant_topic_string"]
        + " - "
        + dta["percent_dominant_topic_string"]
        + " - "
        + dta["id"]
        + " - "
        + dta["text_no_newline"]
    )
    output_string = dta.output_string.str.cat(sep="\n\n")
    dtxt_path.write_text(output_string)


def create_top_sample_dominant_topic_text_file(dta, dtst_path):
    dta["text_no_newline"] = dta["text"].str.replace("\n", " |NEWLINE_CHAR| ")
    dta = dta.sort_values(
        ["dominant_topic_string", "percent_dominant_topic_string"], ascending=[True, False]
    )
    dta = dta.groupby("dominant_topic_string").head(4).reset_index(drop=True)
    dta["output_string"] = (
        dta["dominant_topic_string"]
        + " - "
        + dta["percent_dominant_topic_string"]
        + " - "
        + dta["id"]
        + " - "
        + dta["text_no_newline"]
    )
    output_string = dta.output_string.str.cat(sep="\n\n")
    dtst_path.write_text(output_string)


def create_output_string(dta):
    dta["output_string"] = (
        dta["id"]
        + " - "
        + dta["year_month"]
        + " - "
        + dta["comment"]
        + " - "
        + dta["text_combined"]
    )
    return dta.output_string.str.cat(sep="\n")


@timer
def load_lda_model(ldam_path):
    model_bytes = ldam_path.read_bytes()
    return tp.LDAModel.loads(model_bytes)


def create_lda_output_files(
    target_label: str, topic_model_path: Path, dta: pd.DataFrame, num_topics: int
):
    topic_model_num_path = topic_model_path / f"{target_label}_{str(num_topics).zfill(4)}_topics"
    topic_model_num_path.mkdir(parents=True, exist_ok=True)
    target_num_label = f"{target_label}_{str(num_topics).zfill(4)}"
    ldam_path = topic_model_num_path / f"{target_num_label}_lda_model.ldam"
    mdl = load_lda_model(ldam_path)
    pyvs_path = topic_model_num_path / f"{target_num_label}_pyLDAvis.html"
    create_pyLDAVis(mdl, pyvs_path)
    topt_path = topic_model_num_path / f"{target_num_label}_topics_terms.txt"
    create_topic_summery_text_page(mdl, topt_path)

    # domt_path = topic_model_num_path / f"{target_num_label}_dominate_topics.parquet"
    # domt_path = topic_model_num_path / f"{target_num_label}_dominate_topics.parquet"
    # dtxt_path = topic_model_num_path / f"{target_num_label}_dominate_topics.txt"
    # dtst_path = topic_model_num_path / f"{target_num_label}_dominate_topics_sample.txt"
    # tpex_path = topic_model_num_path / f"{target_num_label}_dominant_topics_terms.xlsx"
    # tgex_path = topic_model_num_path / f"{target_num_label}_pub_count_by_year.xlsx"
    ltxp_path = topic_model_path / f"{target_label}_tomotopy_lemmatized_texts.parquet"
    dtdx_path = topic_model_num_path / f"{target_num_label}_dominant_topic_by_doc.xlsx"
    dtcx_path = topic_model_num_path / f"{target_num_label}_dominant_topic_count.xlsx"
    dtdp_path = topic_model_num_path / f"{target_num_label}_dominant_topic_by_doc.parquet"
    corp_path = topic_model_path / f"{target_label}_corpus.tomotopy"
    extract_dominant_topic(mdl, ltxp_path, dtdx_path, corp_path, dtdp_path, dtcx_path)
    # dta = create_dominant_topic_dta(domt_path, dta, text_dom_topics, percents_dom_topic)
    # create_dominant_topic_text_file(dta, dtxt_path, tpex_path, tgex_path)
    # create_top_sample_dominant_topic_text_file(dta, dtst_path)



# --------------------------------------------------------------------------------------
# File: "topic_model_002_lda.py"
# Dir: "topic_model_scripts\topic_model_002_lda.py"
# Created: 2021-08-17
# --------------------------------------------------------------------------------------

"""
this file is meant to create a topic model using the HDP algorithm
I used  Normalized  Pointwise Mutual Information (NPMI) for model evaluation
"NPMI ranges from [-1,1] and it measures how much the top-10 words of a topic are related to each other, where higher positive NMPI is better."
BERT for Arabic Topic Modeling: An Experimental Study on BERTopic Technique by Abeer Abuzayed Hend Al-Khalifa
"""

from SandsPythonFunctions.TimerFunctions import function_timer as timer
from pathlib import Path
from tqdm import tqdm
import msgpack
import pretty_errors
import tomotopy as tp


def pre_lda_model_trianing(num_topics, corp_path):
    corpus = tp.utils.Corpus.load(corp_path)
    mdl = tp.LDAModel(k=num_topics, min_cf=10, min_df=5, corpus=corpus)
    mdl.train(0)
    min_cf_5_per = int(len(mdl.used_vocabs) * 0.05)
    min_df_5_per = int(len(mdl.docs) * 0.05)
    rm_top_5_per = int(20)
    return corpus, min_cf_5_per, min_df_5_per, rm_top_5_per


def make_hyperparameter_combos():
    term_weights = [tp.TermWeight.ONE, tp.TermWeight.IDF, tp.TermWeight.PMI]
    alphas = [0.01, 0.1, 0.5, 1.0]
    etas = [0.01, 0.1, 0.5, 1.0]
    hyperparameters_combos = []
    for alpha in alphas:
        for eta in etas:
            for term_weight in term_weights:
                hyperparameters_combos.append([term_weight, alpha, eta])
    return hyperparameters_combos


def train_lda_model(
    corpus,
    num_topics,
    min_cf_5_per,
    min_df_5_per,
    rm_top_5_per,
    hyperparameters=[tp.TermWeight.ONE, 0.1, 0.01],
):
    mdl = tp.LDAModel(
        min_cf=min_cf_5_per,
        min_df=min_df_5_per,
        corpus=corpus,
        rm_top=rm_top_5_per,
        k=num_topics,
        tw=hyperparameters[0],
        alpha=hyperparameters[1],
        eta=hyperparameters[2],
    )
    mdl.train(0)
    # print(
    #     f"Training LDA model with num_topics ({num_topics}), term_weight ({str(hyperparameters[0])}), alpha ({hyperparameters[1]}) and eta ({hyperparameters[2]})"
    # )
    # for i in range(0, 1000, 500):  # TESTCODE:
    #     mdl.train(500)  # TESTCODE:
    #     print(f"Iteration: {i}\tLog-likelihood: {mdl.ll_per_word}")  # TESTCODE:
    return mdl


def coherence_scores_file_manage(cohs_path, topics_coherence):
    import json

    all_coherence_scores = []
    if cohs_path.exists():
        with open(cohs_path, "r") as output_json:
            all_coherence_scores = json.load(output_json)
        # all_coherence_scores.append(old_topics_coherence)
    all_coherence_scores.append(topics_coherence)
    with open(cohs_path, "w") as output_json:
        json.dump(all_coherence_scores, output_json)


def calculate_coherence_scores(num_topics, hyperparameters, mdl):
    coh_u_mass = tp.coherence.Coherence(mdl, coherence="u_mass")
    coh_c_uci = tp.coherence.Coherence(mdl, coherence="c_uci")
    coh_c_v = tp.coherence.Coherence(mdl, coherence="c_v")
    coh_c_npmi = tp.coherence.Coherence(mdl, coherence="c_npmi")
    ave_coherence_u_mass = coh_u_mass.get_score()
    ave_coherence_c_uci = coh_c_uci.get_score()
    ave_coherence_c_v = coh_c_v.get_score()
    ave_coherence_c_npmi = coh_c_npmi.get_score()
    return [
        num_topics,
        str(hyperparameters[0]),
        hyperparameters[1],
        hyperparameters[2],
        ave_coherence_u_mass,
        ave_coherence_c_uci,
        ave_coherence_c_v,
        ave_coherence_c_npmi,
    ]


@timer
def test_lda_parameters(ldam_path, corp_path, cohs_path, num_topics, hyper_parameter_testing):
    if not ldam_path.exists():
        print(f"Training ldamodel for {str(num_topics).zfill(4)} topics")
        corpus, min_cf_5_per, min_df_5_per, rm_top_5_per = pre_lda_model_trianing(
            num_topics, corp_path
        )
        mdl = tp.LDAModel(k=num_topics, min_cf=10, min_df=5, corpus=corpus)
        if hyper_parameter_testing:
            hyperparameter_combos = make_hyperparameter_combos()
            for hyperparameters in tqdm(
                hyperparameter_combos, desc=f"Testing hyperparameters for {num_topics} topics"
            ):
                mdl = train_lda_model(
                    corpus, num_topics, min_cf_5_per, min_df_5_per, rm_top_5_per, hyperparameters
                )
                topics_coherence = calculate_coherence_scores(num_topics, hyperparameters, mdl)
                coherence_scores_file_manage(cohs_path, topics_coherence)
        else:
            hyperparameters = [tp.TermWeight.ONE, 0.1, 0.01]
            mdl = train_lda_model(
                corpus, num_topics, min_cf_5_per, min_df_5_per, rm_top_5_per, hyperparameters
            )
            topics_coherence = calculate_coherence_scores(num_topics, hyperparameters, mdl)
            coherence_scores_file_manage(cohs_path, topics_coherence)
        # mdl.summary()
        # print("Writing model file")
        # model_bytes = mdl.saves(full=True)
        # ldam_path.write_bytes(model_bytes)


def create_lda_model(
    ldam_path, corp_path, cohs_path, num_topics, hyperparameters=[tp.TermWeight.ONE, 0.1, 0.01]
):
    corpus, min_cf_5_per, min_df_5_per, rm_top_5_per = pre_lda_model_trianing(num_topics, corp_path)
    mdl = train_lda_model(
        corpus, num_topics, min_cf_5_per, min_df_5_per, rm_top_5_per, hyperparameters
    )
    topics_coherence = calculate_coherence_scores(num_topics, hyperparameters, mdl)
    coherence_scores_file_manage(cohs_path, topics_coherence)
    model_bytes = mdl.saves(full=True)
    ldam_path.write_bytes(model_bytes)


@timer
def get_coherence_model_score(num_topics, ldam_path, cohs_path):
    """There is a bug with the c_v method this issue has been reported here: https://github.com/bab2min/tomotopy/issues/126"""
    import json

    if not cohs_path.exists():
        print("Processing coherence score")
        model_bytes = ldam_path.read_bytes()
        mdl = tp.LDAModel.loads(model_bytes)
        coh_u_mass = tp.coherence.Coherence(mdl, coherence="u_mass")
        coh_c_uci = tp.coherence.Coherence(mdl, coherence="c_uci")
        coh_c_v = tp.coherence.Coherence(mdl, coherence="c_v")
        coh_c_npmi = tp.coherence.Coherence(mdl, coherence="c_npmi")
        ave_coherence_u_mass = coh_u_mass.get_score()
        ave_coherence_c_uci = coh_c_uci.get_score()
        ave_coherence_c_v = coh_c_v.get_score()
        ave_coherence_c_npmi = coh_c_npmi.get_score()
        topics_coherence = [
            num_topics,
            ave_coherence_u_mass,
            ave_coherence_c_uci,
            ave_coherence_c_v,
            ave_coherence_c_npmi,
        ]
        with open(cohs_path, "w") as output_json:
            json.dump(topics_coherence, output_json)
    else:
        with open(cohs_path, "r") as output_json:
            topics_coherence = json.load(output_json)
    print(
        f"The coherence for {str(topics_coherence[0]).zfill(4)} topics: {round(topics_coherence[1], 3)}"
    )
    return topics_coherence


def create_topic_model(
    target_label: str,
    topic_model_path: Path,
    num_topics: int,
    hyperparameter_testing=False,
    hyperparameter_override_list=False,
):
    topic_model_num_path = topic_model_path / f"{target_label}_{str(num_topics).zfill(4)}_topics"
    target_num_label = f"{target_label}_{str(num_topics).zfill(4)}"
    corp_path = topic_model_path / f"{target_label}_corpus.tomotopy"
    ldam_path = topic_model_num_path / f"{target_num_label}_lda_model.ldam"
    cohs_path = topic_model_path / f"{target_label}_lda_model_coherence.json"
    if hyperparameter_testing:
        test_lda_parameters(ldam_path, corp_path, cohs_path, num_topics, hyperparameter_testing)
    elif hyperparameter_override_list:
        topic_model_num_path.mkdir(parents=True, exist_ok=True)
        create_lda_model(ldam_path, corp_path, cohs_path, num_topics, hyperparameter_override_list)
    else:
        topic_model_num_path.mkdir(parents=True, exist_ok=True)
        create_lda_model(ldam_path, corp_path, num_topics)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sandsldahelper']

package_data = \
{'': ['*']}

install_requires = \
['flake8>=3.9.2,<4.0.0',
 'isort>=5.9.3,<6.0.0',
 'msgpack>=1.0.2,<2.0.0',
 'nltk>=3.6.2,<4.0.0',
 'numpy>=1.21.2,<2.0.0',
 'openpyxl>=3.0.8,<4.0.0',
 'pandas>=1.3.3,<2.0.0',
 'pretty-errors>=1.2.24,<2.0.0',
 'pyLDAvis>=3.3.1,<4.0.0',
 'pyarrow>=5.0.0,<6.0.0',
 'sandspythonfunctions>=0.1.0,<0.2.0',
 'spacy>=3.1.2,<4.0.0',
 'tomotopy>=0.12.2,<0.13.0',
 'tqdm>=4.62.2,<5.0.0',
 'vaex>=4.5.0,<5.0.0',
 'wheel>=0.37.0,<0.38.0',
 'zstandard>=0.15.2,<0.16.0']

setup_kwargs = {
    'name': 'sandsldahelper',
    'version': '0.1.0',
    'description': 'This python package is meant to make creating, testing and creating useful output files for LDA topic modeling much easier. It uses Tomotopy for the topic modeling.',
    'long_description': None,
    'author': 'Levi Sands',
    'author_email': 'ldsands@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



# --------------------------------------------------------------------------------------
# File: "0420_topic_modeling_prep.py"
# Dir: "scripts"
# Created: 2021-01-27
# --------------------------------------------------------------------------------------

"""
this file is create a gensim topic model based on factors with the highest coherence value

the file that cleans the text is required before this script is run

KEY: this file requires the installation of message pack which is a compacted version of json

you can install message pack by using this line to install the package

`pip install msgpack`
"""

from SandsPythonFunctions.TimerFunctions import function_timer as timer
from pathlib import Path
from tqdm import tqdm
import SandsPythonFunctions.ParquetFunctions as pf
import SandsPythonFunctions.TextFunctions as tf
import msgpack
import numpy as np
import pandas as pd
import pretty_errors
import tomotopy as tp


def get_files(topic_model_path, target_label: str):
    """
    gets the path for the script location and returns the parquet files located in the
    local folder
    """
    topic_model_path.mkdir(parents=True, exist_ok=True)
    return topic_model_path


def get_stopwords():
    """
    this will test to see if the stopwords from the nltk module have already been
    downloaded if they have not they will be download this function is needed for both
    word embedding and topic modeling and is just overall useful
    """
    from nltk.downloader import download
    from nltk.corpus import stopwords

    try:
        return stopwords.words("english")
    except:
        print(f"NLTK needs to download the stopwords. This will take a while.")
        download("stopwords")
        print(f"NLTK has finished downloading stopwords.")
        return stopwords.words("english")


@timer
def remove_stopwords_function(dta):
    """
    this function removes all of the stopwords from every item in a column in this case
    in the cleaned column
    this function is needed for both word embedding and topic modeling and is just
    overall useful
    """
    stop_words = get_stopwords()
    dta["text"] = dta["text"].apply(
        lambda x: " ".join([word for word in x.split() if word not in (stop_words)])
    )
    return dta


@timer
def lemm_text_prep(dta):
    """this will clean the text of the column name given"""
    import vaex

    print("Dropping missing values using pandas")  # TESTCODE:
    dta = dta.dropna(subset=["text"])
    # dta["text"] = dta["text"].str.replace({"nan": "", "NaN": "", "None": "", "none": "", None: ""})
    dta["text"] = dta["text"].str.replace(
        r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+",
        " ",
        regex=True,
    )
    print("Processing removal of non-alpha characters")  # TESTCODE:
    dta["text"] = dta["text"].apply(
        lambda x: " ".join([word for word in x.split() if word.isalpha()])
    )
    print("Done processing removal of non-alpha characters")  # TESTCODE:
    print("Converting to Vaex Dataframe")  # TESTCODE:
    dta = vaex.from_pandas(dta)
    print("Done converting to Vaex Dataframe")  # TESTCODE:
    print("Now processing using vaex")  # TESTCODE:
    dta["text"] = dta["text"].str.lower()
    dta["text"] = dta["text"].str.replace("[^\w\s]", "", regex=True)  # punctuation
    dta["text"] = dta["text"].str.replace(r"'", " ", regex=True)  # apostrophe
    dta["text"] = dta["text"].str.replace(r"<[^<>]+(>|$)", " ", regex=True)
    dta["text"] = dta["text"].str.replace(r"<img[^<>]+(>|$)", " ", regex=True)
    dta["text"] = dta["text"].str.replace(r"\[img_assist[^]]*?\]", " ", regex=True)
    dta["text"] = dta["text"].str.replace(r"\W+", " ", regex=True)
    dta["text"] = dta["text"].str.replace(r" +", " ", regex=True)
    # dta = tf.clean_text_columns(
    #     dta,
    #     target_column="text",
    #     remove_punctuation=True,
    #     make_lower=False,
    #     remove_stopwords=True,
    #     stem_strings=False,
    #     remove_urls=True,
    #     get_number_of_words=False,
    #     remove_empty_body_posts=True,
    #     remove_non_alpha=True,
    #     print_timings=False,
    # )
    print("Converting back to pandas (and processing the 8 regex string functions)")  # TESTCODE:
    dta = dta.to_pandas_df()
    print("Done converting back to pandas")  # TESTCODE:
    print("Processing the decoding of ascii characters (using pandas)")  # TESTCODE:
    dta["text"] = dta["text"].str.encode("ascii", "ignore").str.decode("ascii")
    print("Done processing the decoding of ascii characters (using pandas)")  # TESTCODE:
    return dta


@timer
def lemmatization_spacy_joblib(texts):
    """this lemmatizes the text
    Make sure to install joblib by using this command `pip install joblib`
    """
    import sys
    import spacy
    from joblib import Parallel, delayed

    try:
        # nlp = spacy.load("en_core_web_sm", disable=["parser", "ner"], n_threads=6)
        # nlp = spacy.load("en_core_web_sm", disable=["tagger", "parser", "ner"])
        nlp = spacy.load("en_core_web_sm", disable=["parser", "ner"])
        # nlp.add_pipe(nlp.create_pipe("sentencizer"))
        nlp.add_pipe("sentencizer")
    except OSError:
        sys.exit(
            "You must run `python -m spacy download en_core_web_sm` to download data required to run spacy"
        )

    def lemmatize_pipe(doc):
        allowed_postags = ["NOUN", "ADJ", "VERB", "ADV"]
        return [
            str(token.lemma_).lower()
            for token in doc
            if token.is_alpha and token.text.lower() and token.pos_ in allowed_postags
        ]

    def chunker(iterable, total_length, chunksize):
        return (iterable[pos : pos + chunksize] for pos in range(0, total_length, chunksize))

    def flatten(list_of_lists):
        "Flatten a list of lists to a combined list"
        return [item for sublist in list_of_lists for item in sublist]

    def process_chunk(texts):
        return [lemmatize_pipe(doc) for doc in nlp.pipe(texts, batch_size=20)]

    def preprocess_parallel(texts, chunksize=100):

        executor = Parallel(n_jobs=10)
        do = delayed(process_chunk)
        tasks = (do(chunk) for chunk in chunker(texts, len(texts), chunksize=chunksize))
        result = executor(tasks)
        return flatten(result)

    return preprocess_parallel(texts, chunksize=3500)


@timer
def lemmatize_text(dta):
    dta = lemm_text_prep(dta)
    dta = remove_stopwords_function(dta)
    texts = dta["text"].tolist()
    lemmatized_texts = lemmatization_spacy_joblib(texts)
    dta["lemmatized_texts"] = lemmatized_texts
    dta[["id", "text", "lemmatized_texts"]]
    return dta


@timer
def load_lemmatized_texts(dta, ltxp_path):
    dta = lemmatize_text(dta)
    dta["lemmatized_texts"] = [" ".join(text) for text in dta["lemmatized_texts"].values.tolist()]
    dta = dta[dta["lemmatized_texts"] != ""]
    dta = tf.remove_words_below_freq(dta, "lemmatized_texts", min_num_word_freq=5)
    dta = dta[dta["num_words"] > 4]
    pf.save_dataframe_as_parquet(dta, ltxp_path)
    return dta["lemmatized_texts"].values.tolist()


@timer
def prepare_lemmatized_texts(topic_model_path, target_label, dta, lmtx_path, ltxp_path):
    """TODO: Docstring"""
    import SandsPythonFunctions.ParquetFunctions as pf

    if lmtx_path.exists():
        return msgpack.unpackb(lmtx_path.read_bytes())
        # return msgpack.load(open(lmtx_path, "rb"))
    else:
        lemmatized_texts = load_lemmatized_texts(dta, ltxp_path)
        lmtx_path.write_bytes(msgpack.packb(lemmatized_texts))
        # msgpack.dump(lemmatized_texts, open(lmtx_path, "wb"))
        return lemmatized_texts


@timer
def extract_ngrams(corpus: tp.utils.Corpus, lemmatized_texts: list, ngrm_path: Path):
    """takes the corpus and finds the most common ngrams more information can be found at the two sites below:
    https://bab2min.github.io/tomotopy/v0.12.2/en/utils.html#tomotopy.utils.Corpus.extract_ngrams
    https://bab2min.github.io/tomotopy/v0.12.2/en/utils.html#tomotopy.utils.Corpus.concat_ngrams
    TODO: figure out how you should use the length of the lemmatized_texts to influance the ngram variables
    """
    num_docs_5_per = int(len(lemmatized_texts) * 0.05)
    cands = corpus.extract_ngrams(
        min_cf=20,
        min_df=10,
        max_len=5,
        max_cand=5000,
        normalized=True,
        min_score=0.5,
    )
    print(f"There are {len(cands)} ngram found")  # TESTCODE:
    # for cand in cands: print(cand)  # TESTCODE:
    output_string = ""
    for cand in cands:
        output_string = f"{output_string}cand\n\n"
    ngrm_path.write_text(output_string)
    corpus.concat_ngrams(cands, delimiter="_")
    return corpus


@timer
def create_corpus(lemmatized_texts: list, corp_path: Path, ltxp_path: Path, ngrm_path: Path):
    """this takes the list of lemmatized texts and tokenizes, extracts ngrams, creates saves a tomotopy corpus

    Args:
        lemmatized_texts (list): a list of strings which are the lemmatized texts
        corp_path (Path): corpus pathlib.Path
    """
    if not corp_path.exists():
        corpus = tp.utils.Corpus(tokenizer=tp.utils.SimpleTokenizer(stemmer=None))
        corpus.process(lemmatized_texts)
        corpus = extract_ngrams(corpus, lemmatized_texts, ngrm_path)
        corpus.save(corp_path)


@timer
def prep_topic_model_files(target_label: str, topic_model_path: Path, dta: pd.DataFrame):
    lmtx_path = topic_model_path / f"{target_label}_tomotopy_lemmatized_texts.msgpack"
    ltxp_path = topic_model_path / f"{target_label}_tomotopy_lemmatized_texts.parquet"
    corp_path = topic_model_path / f"{target_label}_corpus.tomotopy"
    ngrm_path = topic_model_path / f"{target_label}_ngram.txt"
    topic_model_path = get_files(topic_model_path, target_label)
    lemmatized_texts = prepare_lemmatized_texts(
        topic_model_path, target_label, dta, lmtx_path, ltxp_path
    )
    create_corpus(lemmatized_texts, corp_path, ltxp_path, ngrm_path)
